// Copyright 2025, IPng Networks GmbH, Pim van Pelt <pim@ipng.ch>

package vpp

import (
	"os"

	"go.fd.io/govpp/api"
	interfaces "go.fd.io/govpp/binapi/interface"
	"go.fd.io/govpp/binapi/interface_types"

	"govpp-snmp-agentx/logger"
)

// InterfaceDetails holds detailed information about a VPP interface
type InterfaceDetails struct {
	SwIfIndex     interface_types.InterfaceIndex
	InterfaceName string
	MacAddress    []byte
	Speed         uint64
	AdminStatus   bool
	OperStatus    bool
	MTU           uint32
}

// InterfaceEventCallback is called when interface events occur
type InterfaceEventCallback func(details []InterfaceDetails)

// GetAllInterfaceDetails retrieves detailed information for all interfaces
func GetAllInterfaceDetails(ch api.Channel) ([]InterfaceDetails, error) {
	logger.Debugf("Retrieving all interface details from VPP")

	// Get all interfaces
	reqCtx := ch.SendMultiRequest(&interfaces.SwInterfaceDump{
		SwIfIndex: ^interface_types.InterfaceIndex(0), // All interfaces
	})

	var details []InterfaceDetails

	for {
		iface := &interfaces.SwInterfaceDetails{}
		stop, err := reqCtx.ReceiveReply(iface)
		if stop {
			break
		}
		if err != nil {
			logger.Debugf("Error retrieving interface details: %v", err)
			return nil, err
		}

		// Convert VPP interface flags to admin/oper status
		adminUp := (iface.Flags & interface_types.IF_STATUS_API_FLAG_ADMIN_UP) != 0
		operUp := (iface.Flags & interface_types.IF_STATUS_API_FLAG_LINK_UP) != 0

		detail := InterfaceDetails{
			SwIfIndex:     iface.SwIfIndex,
			InterfaceName: string(iface.InterfaceName),
			MacAddress:    iface.L2Address[:],
			Speed:         uint64(iface.LinkSpeed) * 1000, // Convert Kbps to bps
			AdminStatus:   adminUp,
			OperStatus:    operUp,
			MTU:           uint32(iface.LinkMtu),
		}

		details = append(details, detail)

		logger.Debugf("Interface %d (%s): MAC=%x, Speed=%d, Admin=%t, Oper=%t, MTU=%d",
			detail.SwIfIndex, detail.InterfaceName, detail.MacAddress,
			detail.Speed, detail.AdminStatus, detail.OperStatus, detail.MTU)
	}

	logger.Debugf("Retrieved details for %d interfaces", len(details))
	return details, nil
}

func WatchInterfaceEvents(ch api.Channel, callback InterfaceEventCallback) error {
	logger.Debugf("WatchInterfaceEvents() called - starting interface event monitoring")

	notifChan := make(chan api.Message, 100)

	// subscribe for specific event message
	logger.Debugf("Subscribing to interface events...")
	sub, err := ch.SubscribeNotification(notifChan, &interfaces.SwInterfaceEvent{})
	if err != nil {
		logger.Debugf("error subscribing to interface events: %v", err)
		return err
	}
	logger.Debugf("Successfully subscribed to interface events")

	// enable interface events in VPP
	logger.Debugf("Enabling interface events in VPP...")
	err = ch.SendRequest(&interfaces.WantInterfaceEvents{
		PID:           uint32(os.Getpid()),
		EnableDisable: 1,
	}).ReceiveReply(&interfaces.WantInterfaceEventsReply{})
	if err != nil {
		logger.Debugf("error enabling interface events: %v", err)
		return err
	}

	logger.Debugf("Interface events enabled in VPP, starting event listener goroutine")

	// receive notifications
	go func() {
		logger.Debugf("Interface event listener goroutine started")
		defer func() {
			logger.Debugf("Interface event listener goroutine shutting down")
			// disable interface events in VPP
			err = ch.SendRequest(&interfaces.WantInterfaceEvents{
				PID:           uint32(os.Getpid()),
				EnableDisable: 0,
			}).ReceiveReply(&interfaces.WantInterfaceEventsReply{})
			if err != nil {
				logger.Debugf("error disabling interface events: %v", err)
			}

			// unsubscribe from receiving events
			err = sub.Unsubscribe()
			if err != nil {
				logger.Debugf("error unsubscribing from interface events: %v", err)
			}
		}()

		logger.Debugf("Interface event listener waiting for events...")
		for notif := range notifChan {
			e := notif.(*interfaces.SwInterfaceEvent)
			logger.Debugf("interface event: SwIfIndex=%d, Flags=%d, Deleted=%t",
				e.SwIfIndex, e.Flags, e.Deleted)

			// When an interface event occurs, retrieve all interface details and call callback
			if callback != nil {
				details, err := GetAllInterfaceDetails(ch)
				if err != nil {
					logger.Debugf("Failed to retrieve interface details after event: %v", err)
				} else {
					logger.Debugf("Calling interface event callback with %d interfaces", len(details))
					callback(details)
				}
			}
		}
		logger.Debugf("Interface event listener goroutine ended")
	}()

	return nil
}
